'use strict';

exports.__esModule = true;
exports.SubRouter = exports.MockerRouter = undefined;

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _pathToRegexp = require('path-to-regexp');

var _pathToRegexp2 = _interopRequireDefault(_pathToRegexp);

var _utils = require('../utils/');

var _request3 = require('./request');

var _response = require('./response');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var routerLog = _utils.debug.scope('router');

// bacic HTTP request methods in fetch standard, see
// https://fetch.spec.whatwg.org/#concept-method
var methods = ['get', 'post', 'put', 'head', 'delete', 'options'];

var MockerRouter = exports.MockerRouter = function () {

  /**
   * Constructs a new router object
   *
   * @param {string} [baseURL='/'] The base url of this router
   * @param {Array<MiddlewareFn>} [middleware=[]] Middleware that are attached to this router
   */


  /**
   * A collection of all routings
   *
   * @private
   * @type {Array<RouteRule>}
   */


  /**
   * The origin which this router belongs to
   *
   * @private
   * @type {string}
   */

  /**
   * save all routers for lazy evaluations
   *
   * @static
   * @type {Array<MockerRouter>}
   */
  function MockerRouter() {
    var _middleware;

    var baseURL = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : '/';
    var middleware = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];

    _classCallCheck(this, MockerRouter);

    this.baseURL = undefined;
    this._origin = undefined;
    this._basePath = undefined;
    this._rules = [];
    this._middleware = [];

    MockerRouter.routers.push(this);

    (_middleware = this._middleware).push.apply(_middleware, middleware);

    // resolve url based on current origin for relative path
    // `location.origin` is not supported in IE
    var url = new URL(baseURL, location.href);

    this._origin = url.origin;
    this._basePath = url.pathname.replace(/\/$/, ''); // remove trailing slash

    this.baseURL = this._origin + this._basePath;
  }

  /**
   * Create a new router with the given path as scope
   *
   * @param {string}        [path]
   * @return {MockerRouter}
   */


  /**
   * A collection of all middleware
   *
   * @private
   * @type {Array<MiddlewareFn>}
   */


  /**
   * The base path of this router
   *
   * @private
   * @type {string}
   */


  /**
   * A parsed base URL
   *
   * @readonly
   * @type {string}
   */


  MockerRouter.prototype.scope = function scope(path) {
    // in case of falsy values
    if (!path) {
      path = '/';
    }

    if (path[0] !== '/') {
      throw new TypeError('the scope of router should start with "/", got ' + path);
    }

    return new MockerRouter(this.baseURL + path, this._middleware);
  };

  /* istanbul ignore next */


  MockerRouter.prototype.base = function base(path) {
    routerLog.warn('`router.base()` is deprecated, use `router.scope()` instead.');

    return this.scope(path);
  };

  /**
   * Create a scoped router with the given path as
   * route path for every routing method.
   *
   * @param  {string|RegExp} path Routing's path
   * @return {SubRouter}
   */


  MockerRouter.prototype.route = function route(path) {
    return new SubRouter(this, path);
  };

  /**
   * Attach middleware to current router
   *
   * @param  {MiddlewareFn} fn Middleware function
   * @return {this}
   */


  MockerRouter.prototype.use = function use(fn) {
    if (typeof fn !== 'function') {
      throw new TypeError('Middleware is expected to be a function');
    }

    this._middleware.push(fn);

    return this;
  };

  /**
   * Register a new routing to current router
   *
   * @private
   * @param  {string}       method   HTTP method
   * @param  {RoutePath}    path     Routing path rule
   * @param  {function|any} callback Routing callback handler
   * @return {this}
   */


  MockerRouter.prototype._register = function _register(method, path, callback) {
    method = method.toUpperCase();

    var regex = (0, _pathToRegexp2.default)(path);

    var cb = void 0;

    if (typeof callback === 'function') {
      cb = callback;
    } else {
      // shorthand method
      cb = function cb(_request, response) {
        response.send(callback);
      };
    }

    this._rules.push({
      method: method,
      path: path,
      regex: regex,
      callback: cb,
      keys: regex.keys,
      baseURL: this.baseURL,
      isAll: method === 'ALL'
    });

    return this;
  };

  /**
   * Match the proper routing, return `true` if rule matched
   *
   * @private
   * @param {FetchEvent} event Fetch event
   * @return {boolean}
   */


  MockerRouter.prototype._match = function _match(event) {
    var _this = this;

    var request = event.request;

    // `request.url` maybe relative in legacy mode

    var url = new URL(request.url, location.href);

    if (url.origin !== this._origin) {
      return false;
    }

    // strip router's base path
    var re = new RegExp('^' + this._basePath);
    var path = url.pathname.replace(re, '');

    var _loop = function _loop(rule) {
      var method = rule.method,
          regex = rule.regex,
          callback = rule.callback;


      if (regex.test(path) && (request.method === method || rule.isAll)) {
        //! Response object must be constructed synchronously
        var _request2 = new _request3.MockerRequest(event, rule);
        var response = new _response.MockerResponse(event);

        // apply (async) middleware
        Promise.all(_this._middleware.map(function (fn) {
          return fn.call(event, _request2, response);
        })).then(function () {
          callback.call(event, _request2, response);
        });

        return {
          v: true
        };
      }
    };

    for (var _iterator = this._rules, _isArray = Array.isArray(_iterator), _i = 0, _iterator = _isArray ? _iterator : _iterator[Symbol.iterator]();;) {
      var _ref;

      if (_isArray) {
        if (_i >= _iterator.length) break;
        _ref = _iterator[_i++];
      } else {
        _i = _iterator.next();
        if (_i.done) break;
        _ref = _i.value;
      }

      var rule = _ref;

      var _ret = _loop(rule);

      if ((typeof _ret === 'undefined' ? 'undefined' : _typeof(_ret)) === "object") return _ret.v;
    }

    return false;
  };

  return MockerRouter;
}();

MockerRouter.routers = [];

var SubRouter = exports.SubRouter = function () {

  /**
   * Constructs a sub router
   * @param {MockerRouter}  router Parent router object
   * @param {string|RegExp} path   Routing path
   */

  /**
   * Router object
   *
   * @private
   * @type {MockerRouter}
   */
  function SubRouter(router, path) {
    _classCallCheck(this, SubRouter);

    this._router = null;
    this._path = undefined;

    this._router = router;
    this._path = path;
  }

  /**
   * Register a new scoped routing
   *
   * @private
   * @param {string}       method   HTTP method
   * @param {function|any} callback Routing callback handler
   * @return {this}
   */


  /**
   * Routing path rule
   *
   * @private
   * @type {string|RegExp}
   */


  SubRouter.prototype._register = function _register(method, callback) {
    this._router._register(method, this._path, callback);

    return this;
  };

  return SubRouter;
}();

var allMethods = ['all'].concat(methods);

// assign all methods to router
allMethods.forEach(function (method) {
  MockerRouter.prototype[method] = function (path, callback) {
    return this._register(method, path, callback);
  };
});

allMethods.forEach(function (method) {
  SubRouter.prototype[method] = function (callback) {
    return this._register(method, callback);
  };
});
//# sourceMappingURL=router.js.map